@@ -12,13 +12,12 @@ use core::{
     convert::Infallible,
     future::Future,
     marker::PhantomData,
+    pin::Pin,
     sync::atomic::{AtomicU32, Ordering::Relaxed},
     task::{Context, Poll, Waker},
 };
 
-use futures::{pin_mut, task::noop_waker};
-
-pub use futures;
+pub use futures::{self, pin_mut};
 pub use nb;
 
 #[cfg(test)]
@@ -74,10 +73,11 @@ pub struct Events<S> {
 /// `Signals` maintains the signal state of an [`Executor`]. This consists of the raised signal set
 /// and the wakeup signal set. Upon polling the future, the raised signal set is frozen to the
 /// then-current value of the event mask, all bits in the wakeup signal set are removed from the
-/// event mask (atomically), and the wakeup signal set is cleared. Current poll functions which are
-/// driven by any of the raised signals will be attempted. If a poll function is not attempted or
-/// does not yet resolve to an output then its signal mask will be OR-ed into the wakeup signal
-/// set. The future won't be polled again until a raised signal matches the wakeup signal set.
+/// event mask (atomically), and the wakeup signal set is cleared. See also [`Step::poll()`]
+/// for behavior when the wakeup signal set is zero. Current poll functions which are driven
+/// by any of the raised signals will be attempted. If a poll function is not attempted or does
+/// not yet resolve to an output then its signal mask will be OR-ed into the wakeup signal set.
+/// The future won't be polled again until a raised signal matches the wakeup signal set.
 ///
 /// The type parameter `S` is an [`EventMask`].
 ///
@@ -94,25 +94,26 @@ pub struct Signals<'a, S> {
 /// A single-threaded, single-future async executor.
 ///
 /// # Typical setup
-/// - First, an event mask is created with [`Events::default()`]. Event masks are Send + Sync`
-///   and shared references to them are enough for all operations, so they work well as `static`
-///   items or other types of shared state. Distribute event mask references to the external
-///   event sources and keep one for the executor.
+/// - First, an event mask is created with [`Events::default()`] (or [`Events::new()`] if
+///   that needs to be `const`). Event masks are Send + Sync` and shared references to the
+///   are enough for all operations, so they work well as `static` items or other types of
+///   shared state. Distribute event mask references to the external event sources and keep
+///   one for the executor.
 ///
-/// - The event mask is then watched with [`Signals::watch()`]. The resulting [`Signals`] is
+/// - The event mask is then watched with [`Events::watch()`]. The resulting [`Signals`] is
 ///   `Send + !Sync`. This means that operations become limited to one thread of execution
 ///   from this point on, so this is usually done in some type of initialization or main
 ///   function instead of in a shared or global context.
 ///
-/// - A new executor is bound with [`Executor::bind()`]. Executors are `!Send + !Sync`:
+/// - A new executor is bound with [`Signals::bind()`]. Executors are `!Send + !Sync`:
 ///   neither it nor the associated `Signals` may escape the current thread. This makes them
 ///   appropriate for construction at the use site.
 ///
 /// - A future is created. It needs a reference to the `Signals` object in order to drive
 ///   poll functions, making it `!Sync` too.
 ///
-/// - Finally, [`Executor::run_with_park()`] blocks and resolves the future while external
-///   event sources direct it through the event mask, possibly with help from the park function.
+/// - Finally, [`Executor::block_on()`] blocks and resolves the future while external event
+///   sources direct it through the event mask, possibly with help from the park function.
 ///
 /// # Examples
 ///
@@ -145,7 +146,7 @@ pub struct Signals<'a, S> {
 /// }
 ///
 /// let events = Arc::new(Events::default());
-/// let signals = Signals::watch(&events);
+/// let signals = events.watch();
 ///
 /// let (tx, rx) = sync_channel(1);
 /// let future = async {
@@ -166,9 +167,14 @@ pub struct Signals<'a, S> {
 ///         events_prod.raise(Ev::QUEUE);
 ///         runner.unpark();
 ///     }
+///
+///     // Notify shutdown
+///     drop(tx);
+///     events_prod.raise(Ev::QUEUE);
+///     runner.unpark();
 /// });
 ///
-/// let result = Executor::bind(&signals).run_with_park(future, |park| {
+/// let result = signals.bind().block_on(future, |park| {
 ///     // thread::park() is event-safe, no lock is required
 ///     let parked = park.race_free();
 ///     if parked.is_idle() {
@@ -180,11 +186,27 @@ pub struct Signals<'a, S> {
 ///
 /// assert_eq!(result, (1..=10).product()); // 3628800
 /// ```
-pub struct Executor<'a, S> {
-    signals: &'a Signals<'a, S>,
+pub struct Executor<'a> {
+    pending: &'a AtomicU32,
+    run: &'a Cell<Run>,
     waker: Waker,
 }
 
+/// A non-blocking executor-future-park state machine.
+///
+/// All executor operations are ultimately implemented with `Step`. It allows fine-grained
+/// control over execution and control flow. `Step` can only perform one poll step at a time
+/// and requires a pinned future. A `Step` object is created by calling [`Executor::step()`].
+///
+/// See [`Executor::block_on()`] for a blocking runner.
+pub struct Step<'exec, 'fut, F, P> {
+    executor: Executor<'exec>,
+    wakeup: u32,
+    last_pending: u32,
+    future: Pin<&'fut mut F>,
+    park: P,
+}
+
 /// A request to park the executor.
 ///
 /// Parking is the mechanism by which the executor *tries* to wait for external event sources when
@@ -196,26 +218,26 @@ pub struct Executor<'a, S> {
 /// # The park protocol
 ///
 /// - First, the executor determines that further progress is unlikely at this moment. The
-///   specifics of this are implementation details that must not be relied upon.
+///   specifics of this process are implementation details that should not be relied upon.
 ///
 /// - The park function is called with a `Park` parameter.
 ///
 /// - The park function enters a context wherein no external events may influence a correct
 ///   decision to sleep or not. For example, a park function that does not sleep at all does
-///   not need to do anything here, since no external event can incorrectly change tha
+///   not need to do anything here, since no external event can incorrectly change that
 ///   behavior. On the other hand, a park function that halts until a hardware interrupt occurs
 ///   would need to enter an interrupt-free context to avoid deadlocks.
 ///
 /// - The park function calls [`Park::race_free()`] while still in the event-safe context.
-///   This produces a [`Parked`] value that demonstrates the call to `race_free()`.
+///   This produces a [`Parked`] value that serves as proof of the call to `race_free()`.
 ///
-/// - If the park function intends to block or sleep then it must first call
-///   [`Parked::is_idle()`] and may be allowed to sleep only if it returns `true`.
+/// - If the park function intends to block or sleep, then it must first call
+///   [`Parked::is_idle()`]. It may be allowed to sleep only if that function returns `true`.
 ///
 /// - If the park function is willing to sleep and is allowed to do so, it must
-///   atomically exit the event-safe context whilst entering the sleep state. A deadlock
-///   is again possible if both operations are not done atomically with respect to each
-///   other.
+///   atomically exit the event-safe context whilst entering the sleep state. A deadlock is
+///   again possible if both operations are not done atomically with respect to each other
+///   (**only for blocking runners, see below**).
 ///
 /// - If the park function sleeps, this state should be automatically exited when an external
 ///   event occurs.
@@ -223,6 +245,17 @@ pub struct Executor<'a, S> {
 /// - The park function returns its [`Parked`] token.
 ///
 /// - The executor resumes.
+///
+/// # Delegating out of the park function
+///
+/// Blocking runners, such as [`Executor::block_on()`], require that the park function's
+/// event-safe context be exited in an atomic manner with respect to the start of whatever
+/// blocking operation. However, this requirement does not hold for [`Step`] **as long as
+/// the park function never exits the event-safe context and it itself never blocks**. Since
+/// [`Step::poll()`] will return after parking, the caller can perform potentially-blocking
+/// operations from the event-safe context. It must still release it atomically if it will
+/// sleep, though. With this technique it is even possible for the `poll()` caller to be an
+/// external event source.
 pub struct Park<'a> {
     pending: &'a AtomicU32,
     wakeup: u32,
@@ -249,6 +282,30 @@ impl<S> Default for Events<S> {
     }
 }
 
+impl<S> Events<S> {
+    /// Construct a new event mask.
+    ///
+    /// This is the same as [`Events::default()`] but is also `const`.
+    pub const fn new() -> Self {
+        Events {
+            events: AtomicU32::new(0),
+            _phantom: PhantomData,
+        }
+    }
+
+    /// Associate a new `Signals` to this event mask.
+    ///
+    /// This indirectly links the executor to the event mask, since creating an [`Executor`]
+    /// requires a `Signals`.
+    #[must_use]
+    pub fn watch(&self) -> Signals<S> {
+        Signals {
+            pending: self,
+            run: Default::default(),
+        }
+    }
+}
+
 impl<S: EventMask> Events<S> {
     /// Raise all events in a mask.
     ///
@@ -260,14 +317,16 @@ impl<S: EventMask> Events<S> {
 }
 
 impl<'a, S: EventMask> Signals<'a, S> {
-    /// Associate a new `Signals` to an event mask.
+    /// Bind a new executor to this signal source.
     ///
-    /// This indirectly links the executor to the event mask, since creating an [`Executor`]
-    /// requires a `Signals`.
-    pub fn watch(pending: &'a Events<S>) -> Self {
-        Signals {
-            pending,
-            run: Default::default(),
+    /// The executor will be constructed with a waker that does nothing. It can be replaced
+    /// by calling [`Executor::with_waker()`].
+    #[must_use]
+    pub fn bind(&self) -> Executor {
+        Executor {
+            pending: &self.pending.events,
+            run: &self.run,
+            waker: futures::task::noop_waker(),
         }
     }
 
@@ -324,29 +383,39 @@ impl<'a, S: EventMask> Signals<'a, S> {
     }
 }
 
-impl<'a, S> Executor<'a, S> {
-    /// Create a new executor with a given signal source.
-    ///
-    /// The executor will be constructed with a waker that does nothing. It can be replaced
-    /// by calling [`Executor::with_waker()`].
-    pub fn bind(signals: &'a Signals<S>) -> Self {
-        Executor {
-            signals,
-            waker: noop_waker(),
-        }
-    }
-
-    /// Replaces the executor's waker with a custom one.
+impl<'exec> Executor<'exec> {
+    /// Replace the executor's waker with a custom one.
     ///
     /// No restrictions are imposed on the waker: `nb-executor` does not use wakers at all.
     /// Application code may define some communication between it and the park function.
-    pub fn with_waker(&mut self, waker: Waker) -> &mut Self {
+    #[must_use]
+    pub fn with_waker(mut self, waker: Waker) -> Self {
         self.waker = waker;
         self
     }
-}
 
-impl<S> Executor<'_, S> {
+    /// Begin stepped execution.
+    ///
+    /// This allows the caller to remain in control of program flow in between polls, unlike
+    /// [`Executor::block_on()`]. The future is run as specified in [`Step`] and [`Signals`].
+    /// `park` is a park function and must follow the [`Park`] protocol. The caller and the
+    /// park function may cooperate to block or sleep outside of the park function within the
+    /// protocol's requirements.
+    pub fn step<'fut, F, P>(self, future: Pin<&'fut mut F>, park: P) -> Step<F, P>
+    where
+        F: Future,
+        P: FnMut(Park) -> Parked,
+        'exec: 'fut,
+    {
+        Step {
+            executor: self,
+            wakeup: 0,
+            last_pending: u32::MAX,
+            future,
+            park,
+        }
+    }
+
     /// Execute a future on this executor, parking when no progress is possible.
     ///
     /// This method will block until the future resolves. There are two possible states of
@@ -361,54 +430,106 @@ impl<S> Executor<'_, S> {
     ///   time. For details, see the parking protocol in [`Park`]. `park` must adhere to
     ///   this protocol.
     ///
-    /// On method enter, the state is set to polling with an all-ones wakeup signal set.
-    /// This ensures that all driven poll functions are called at least once.
-    pub fn run_with_park<F, P>(&mut self, future: F, mut park: P) -> F::Output
+    /// See also [`Step`] and [`Executor::step()`].
+    pub fn block_on<F, P>(self, future: F, park: P) -> F::Output
     where
         F: Future,
         P: FnMut(Park) -> Parked,
     {
         pin_mut!(future);
-
-        let pending = &self.signals.pending.events;
-        let mut cx = Context::from_waker(&self.waker);
-
-        let (mut last_pending, mut wakeup) = (u32::MAX, u32::MAX);
+        let mut step = self.step(future, park);
 
         loop {
-            if last_pending & wakeup != 0 {
-                self.signals.run.set(Run {
-                    raised: last_pending,
-                    wakeup: 0,
-                });
-
-                match future.as_mut().poll(&mut cx) {
-                    Poll::Ready(output) => break output,
-                    Poll::Pending => (),
-                }
-
-                wakeup = self.signals.run.get().wakeup;
-            }
-
-            last_pending = park(Park { pending, wakeup }).last_pending;
-
-            loop {
-                let cleared = last_pending & !wakeup;
-                match pending.compare_exchange_weak(last_pending, cleared, Relaxed, Relaxed) {
-                    Ok(_) => break,
-                    Err(current) => last_pending = current,
-                }
+            match step.poll() {
+                Poll::Pending => continue,
+                Poll::Ready(ready) => break ready,
             }
         }
     }
 
     /// Execute a future in a busy-waiting loop.
     ///
-    /// This is equivalent to calling [`Executor::run_with_park()`] with a park function that
+    /// This is equivalent to calling [`Executor::block_on()`] with a park function that
     /// never sleeps. This is most likely the wrong way to do whatever you intend, prefer to
     /// define a proper wake function.
-    pub fn run_busy<F: Future>(&mut self, future: F) -> F::Output {
-        self.run_with_park(future, |park| park.race_free())
+    pub fn block_busy_on<F: Future>(self, future: F) -> F::Output {
+        self.block_on(future, |park| park.race_free())
+    }
+}
+
+impl<F, P> Step<'_, '_, F, P>
+where
+    F: Future,
+    P: FnMut(Park) -> Parked,
+{
+    /// Perform at most one non-blocking polling attempt.
+    ///
+    /// This method never blocks unless the park function does (or the future, but that's
+    /// totally wrong anyway). The future will be polled given any of these:
+    ///
+    /// - The wakeup signal set is zero. This is intended to provide a reset or "full scan"
+    ///   mechanism. In this case, the raised signal set will be hard-coded to all-ones,
+    ///   regardless of the current value of the event mask. **Note:** A side effect of this
+    ///   special case is that [`core::future::pending()`] and similar constructs might
+    ///   busy-loop depending on the particular park function.
+    ///
+    /// - Any bit in the wakeup signal set matches the event mask. The event mask is
+    ///   atomically frozen at the same time that this condition is check, creating the
+    ///   raised signal set.
+    ///
+    /// The wakeup signal set is cleared just before polling the future (edge-triggering).
+    /// Regardless of the poll result or if the future is polled at all, a park
+    /// operation may be triggered under implementation-defined conditions. If parking
+    /// occurs, it will be the last observable operation before `poll()` returns. The
+    /// return value will be [`Poll::Pending`] if the future is not polled. Do not call
+    /// call `poll()` again after it returns [`Poll::Ready`].
+    ///
+    /// On first call, the wakeup signal set is zero. This ensures that all driven poll
+    /// functions are attempted at least once.
+    pub fn poll(&mut self) -> Poll<F::Output> {
+        let woken = if self.wakeup == 0 {
+            self.last_pending = u32::MAX;
+            true
+        } else {
+            loop {
+                let cleared = self.last_pending & !self.wakeup;
+                let result = self.executor.pending.compare_exchange_weak(
+                    self.last_pending,
+                    cleared,
+                    Relaxed,
+                    Relaxed,
+                );
+
+                match result {
+                    Ok(current) => break cleared != current,
+                    Err(current) => self.last_pending = current,
+                }
+            }
+        };
+
+        if !woken {
+            return Poll::Pending;
+        }
+
+        self.executor.run.set(Run {
+            raised: self.last_pending,
+            wakeup: 0,
+        });
+
+        let mut cx = Context::from_waker(&self.executor.waker);
+        let poll = self.future.as_mut().poll(&mut cx);
+
+        if poll.is_pending() {
+            self.wakeup = self.executor.run.get().wakeup;
+            let park = Park {
+                pending: self.executor.pending,
+                wakeup: self.wakeup,
+            };
+
+            self.last_pending = (self.park)(park).last_pending;
+        }
+
+        poll
     }
 }
 
@@ -432,13 +553,13 @@ impl Parked<'_> {
     /// Check whether useful work is certainly not possible until an event is raised.
     ///
     /// Unlike the calling of the park function, this is not an optimistic operation.
-    /// Its result will be exact as long as the park protocol is correctly followed.
+    /// Its result will be exact as long as the [`Park`] protocol is correctly followed.
     /// A return value of `false` prohibits the park function from sleeping at all:
     /// it should yield control to the executor immediately. A return value of `true`
     /// is a strong hint to sleep, block, or otherwise take over control flow until
     /// some unspecified condition, ideally until an event is raised.
     ///
-    /// See also the park protocol in [`Park`].
+    /// See also the [`Park`] protocol.
     pub fn is_idle(&self) -> bool {
         self.last_pending & self.wakeup == 0
     }
